module Geo.Segmentation

--  ( segment_by_homes
--  , shortest_path
--  , total_flood
--  )
  where

-- $Log: Segmentation.hs,v $
-- Revision 1.28  2002/09/02 15:43:17  mrahn
-- shorter imports
--
-- Revision 1.27  2002/09/02 15:36:15  joe
-- total_flood
--
-- Revision 1.26  2002/09/02 12:41:13  joe
-- shortest_paths

import World.Type
import World.Query.GetFieldSize
import World.Query.Homebases
import World.Query.GetNeighbors
import Geo.Type
import Geo.Util
import Geo.Make
import Geo.Membership
import Util.Debug

import Monad (guard, when)
import IOExts
import IO

total_flood :: (Map, Info) -> Position -> Maybe Int -> IO DistanceMap
-- flood from one point onwards, for max n steps
total_flood ( mappe, info ) start maybe_bound = do
    debug $ unwords [ "total_flood from", show start ]
    let (w,h) =  get_field_size info

    let lohi @ (lo, hi) = bounds mappe
    dist <- newArray lohi $ membership unsegm  (-1) False
    let valid p = lo <= p && p <= hi

    let hh = map (harden w) [ start ]

    -- set initial values (for homes)
    mapM_ ( \ ( base, p ) -> 
	        writeArray dist p $ membership base 0 False
	  ) $ zip [ 1.. ] hh

    flood <- newIORef ( hh :: [ Pos ] )
    accu <- newIORef ( [] :: [ ELP ] )

    flooding True flood accu mappe dist valid w maybe_bound
    debug $ "flooding done"

    return dist



type Flood = [(Set Position)]
type EIWL = [(Edge Int, Way)]

shortest_path :: (Map, Info) -> (Position, Position) -> IO(DistanceMap, Way) 
shortest_path ( mappe, info ) (p, q) = do

    debug $ unwords [ "shortest path from", show p, "to", show q ]

    let homeslist = [ p, q ]
    let (w,h) =  get_field_size info

    let lohi @ (lo, hi) = bounds mappe
    dist <- newArray lohi $ membership unsegm  (-1) False
    let valid p = lo <= p && p <= hi

    let hh = map (harden w) homeslist

    -- set initial values (for homes)
    mapM_ ( \ ( base, p ) -> 
	        writeArray dist p $ membership base 0 False
	  ) $ zip [ 1.. ] hh

    flood <- newIORef ( hh :: [ Pos ] )
    accu <- newIORef ( [] :: [ ELP ] )

    flooding False flood accu mappe dist valid w Nothing
    debug $ "flooding done"

    debug $ unwords [ "shortest path from", show p, "to", show q ]
    elps <- readIORef accu
    let way = case elps of
	 ( e, l, pp ) : _ -> Way { len = l, Geo.Type.pos = soften w pp }
	 _                -> InvalidWay
    debug $ "is  " ++ show way
    return (dist, way)
    
type ELP =  (Edge Int, Int, Pos) 


segment_by_homes :: (Map, Info) -> IO (DistanceMap, WayGraph Int)
segment_by_homes (mappe, info) = do
    homeslist <- homebases (mappe, info)

    let (w,h) =  get_field_size info

    let lohi @ (lo, hi) = bounds mappe
    dist <- newArray lohi $ membership unsegm  (-1) False
    let valid p = lo <= p && p <= hi


    let cut = 10000
    let (pre, post) = splitAt cut homeslist
    debug $ "length homeslist: " ++ show (length pre)
    when ( not $ null post ) $ debug "*** WARNING: homeslist trimmed"

    let hh = map (harden w) pre

    -- set initial values (for homes)
    mapM_ ( \ ( base, p ) -> 
	        writeArray dist p $ membership base 0 False 
	  ) $ zip [ 1.. ] hh


    flood <- newIORef ( hh :: [ Pos ] )
    accu <- newIORef ( [] :: [ ELP ] )

    flooding True flood accu mappe dist valid w Nothing
    debug $ "flooding done"

    elps <- readIORef accu
    let xs = do 
	   ( e, l, p ) <- elps
	   return ( e, Way { len = l, Geo.Type.pos = soften w p } )      

    return (dist, make homeslist xs)



harden :: Int -> Position -> Pos
harden w (x, y) =  ( w + 1 ) * ( y - 1 ) + x 

soften :: Int -> Pos -> Position
soften w p      =  let (di, mo) = divMod (p - 1) (w + 1)
			 in  (mo + 1, di + 1) 

--------------------------------------------------------------------------

flooding complete flood accu mappe dist valid w maybe_bound = do

    -- handle a single position's neighbour 
    -- p itself is already entered in dist
    let handle (p, dir) = do
	   -- debug $ unwords [ "handle for" , show (soften p), show dir ]
	   let q = p + dir
	   -- debug $ unwords [ "neighbour is" , show (soften q) ]
	   when ( valid q ) $ do
		-- debug $ "this is valid"
		x <- readArray mappe q
		when ( ( 46 == x ) || ( 64 == x ) ) $ do
		     -- debug $ "we can go here"
		     mem_p <- readArray dist p
		     let p_base = base_nr mem_p
		     let p_dist = Geo.Membership.distance mem_p
		     mem_q <- readArray dist q
		     let q_base = base_nr mem_q
		     let q_dist = Geo.Membership.distance mem_q
		     if q_base == unsegm
			then do 
			     -- debug $ "free space: occupy"
			     writeArray dist q
			        $ membership p_base (succ p_dist) False
			     modifyIORef  flood ( q : )
			else do
			     -- debug $ unwords [ "already marked" , "base_nr", show q_base ]
			     when ( q_base /= p_base ) $ do
				  -- debug $ "not marked by me"
				  let item = ( kante p_base q_base
					     , q_dist + 1 + p_dist
					     , q
					     )
				  modifyIORef  accu ( item : )



    ps <- readIORef flood -- todo list
    -- debug $ "the waterfront is (hard) " ++ show ps
    -- debug $ "the waterfront is (soft) " ++ show ( map soften ps )
    hPutStr stderr $ "~" ++ show (length ps)
    when ( not $ null ps ) $ do
	 writeIORef flood []   -- init for new todo list
	 mapM_ handle $ do 
		      p <- ps
		      dir <- [ -(w+1) , -1 , w+1, 1 ] -- four directions 
		      return ( p, dir )
	 res <- readIORef accu
	 let more =  flooding complete flood accu mappe dist valid w
	 when ( complete || null res ) $
	      case maybe_bound of
		   Nothing -> more Nothing
		   Just b  -> when ( b > 0 ) $ more (Just $ pred $ b)
	 





--------------------------------------------------------------------------


segment_by_homes_old :: (Map, Info) -> IO(DistanceMap, WayGraph Int)
segment_by_homes_old  (mappe, info) = do
	dist <- create_map info 

	homeslist <- homebases (mappe, info)



	-- create Graph with HomeB
	-- let wayGraph = WayGraph { nodes = mkSet [1.. length (homeslist) ] , edges = emptySet}
	-- start Flood in every HomeB
	let flood0 = map (\ x -> mkSet[x] ) homeslist 
	
	l <- floating (step (mappe, info) dist False) flood0 0 :: IO EIWL
	-- l is of type: [(Edge Int, Way)]
	
	return (dist, make homeslist l)


-- starts two floods from each Position and abort if the floods meeting
shortest_path_old
   :: (Map, Info) -> (Position, Position) -> IO(DistanceMap, Way) 
shortest_path_old (mappe, info) (p1, p2) = do
	dist <- create_map info 
	let flood0 = [mkSet [p1],mkSet [p2]]   
	l <- floating (step (mappe, info) dist True) flood0 0 :: IO EIWL
	-- l is of type: [(Edge Int, Way)]
	if null l then do
		return (dist, Way {len = 0 , Geo.Type.pos = (1,1) })  
	  else do
			let way = snd $ (head l)   
			return (dist, way)
	

create_map:: Info -> IO(DistanceMap)
create_map info = do
	let (width, height) = get_field_size info
	array <- newArray (1, width * height)  
	   $ membership unsegm  (-1) False
  	return array
	
	
floating :: (Flood -> Int -> IO(Flood, EIWL)) -> Flood -> Int ->  IO EIWL
floating st flood i = if isEmptySet$ unionManySets flood then 
						return []
					  else do 
						(flood', l) <- st flood i
						l' <- floating st flood' (i+1)
						return $ l ++ l'   

-- in flood are only new 
step ::(Map, Info) -> DistanceMap -> Bool -> Flood -> Int ->  IO(Flood, EIWL)
step (mappe, info) dist abort f i  = do
	-- insert all Pos of f in DistMap   -> Graph
	l <- insertToDistMap (mappe, info) dist f i
	if and[abort, not $ null l ] then do
		-- abort because
			return ([emptySet], l)
		else do		
		 -- create Succsesors of all in f -> Flood
		 f' <- mapM (get_set_succ (mappe, info) dist) (zip [1..] f)
		 return (f',l)
	
-- write dist and segment information in DistanceMap and return the new WayGraph 
-- to be started with segm=1 and dist = round of flood and WayGraph only nodes 
insertToDistMap :: (Map, Info) -> DistanceMap  ->  Flood -> Int -> IO(EIWL)
insertToDistMap (mappe, info) dist flood i = do 
	ls <- mapM (insertOneSet (mappe, info) dist i) $ zip [1..] flood 	
	return $ concat ls
	 
insertOneSet::(Map, Info) -> DistanceMap -> Int -> (Int, Set Position) -> IO(EIWL)
insertOneSet (mappe, info) dist i (segm, s) = do
	ls <- mapM (insertPos (mappe, info) dist i segm) $ setToList s 
	return $ concat ls
	
insertPos :: (Map, Info) -> DistanceMap -> Int ->  Int -> Position -> IO(EIWL)	
insertPos (_ , info) dist i segm (x,y) = 
	do 
		let (width, _ ) = get_field_size info
		fieldA <- readArray dist (get_index x y width)
		if (base_nr fieldA) == unsegm then do
				writeArray dist (get_index x y width) 
					$ membership segm i False
				return [] 
			else do
				-- found a edge an add it to graph
--				print $ kante (base_nr fieldA) segm
				writeArray dist (get_index x y width) 
					$ membership segm i True 
				return [ (kante (base_nr fieldA) segm
						  , Way { len = i + (Geo.Type.distance fieldA)
							  	, Geo.Type.pos = (x,y)
								}
						 )
					   ] 		   
  

get_set_succ :: (Map, Info) -> DistanceMap ->  (Int,Set Position) -> IO(Set Position) 
get_set_succ (mappe, info) dist (segm,setofpos) = do 
	let	(width, _ ) = get_field_size info
	let list = setToList setofpos
	newlist <- mapM (get_successor (mappe, info) dist width segm) list
	return $ mkSet $ concat newlist 


-- all successors of the Flood segm 
get_successor :: (Map, Info) -> DistanceMap -> Int -> Int -> Position -> IO([Position])
get_successor (mappe, info) dist width segm (x,y) = 
  do
	fieldA <- readArray dist (get_index x y width)
	case (border fieldA) of
		True -> do return []
		False -> do 
                  ns <- get_neighbors (mappe, info) (x,y)
		  list <- mapM (\ (x',y') -> do 
				cont <- readArray dist (get_index x' y' width)
				return ((x',y')
					, (base_nr cont == unsegm)
					 || ( (base_nr cont /= segm) 
						&& ( not $ border cont)))
			       ) ns
		  let flist = do
				  (p, flag) <- list
				  guard flag
				  return p
		  return flist

