module Main.Run ( main ) where

-- $Log: Run.hs,v $
-- Revision 1.24  2002/09/02 17:14:30  joe
-- default server @ icfp1
--
-- Revision 1.23  2002/09/02 14:46:37  joe
-- import FM
--
-- Revision 1.22  2002/09/02 14:41:16  joe
-- highwaysfrom init
--
-- Revision 1.21  2002/09/02 14:30:05  mrahn
-- patti2
--
-- Revision 1.20  2002/09/02 12:15:37  mrahn
-- new pack_update
--
-- Revision 1.19  2002/09/02 11:33:23  joe
-- no debug print
--
-- Revision 1.18  2002/09/02 11:30:04  joe
-- Segmentation durch explizites rumlaufen
--
-- Revision 1.17  2002/09/02 08:46:57  mrahn
-- patti again
--
-- Revision 1.16  2002/09/02 01:18:19  joe
-- tuned
--
-- Revision 1.15  2002/09/01 23:12:20  joe
-- use patti
--
-- Revision 1.14  2002/09/01 19:32:54  joe
-- time fr create
--
-- Revision 1.13  2002/09/01 19:22:56  joe
-- debug msg
--
-- Revision 1.12  2002/09/01 19:19:06  joe
-- timing fixes
--
-- Revision 1.11  2002/09/01 19:09:09  joe
-- time re-org
-- zum anfang Geo fragen
--
-- Revision 1.10  2002/09/01 13:00:48  joe
-- printer
--
-- Revision 1.9  2002/08/31 18:00:06  joe
-- tag
--
-- Revision 1.8  2002/08/31 17:27:01  joe
-- time (vereinfacht)
--
-- Revision 1.7  2002/08/31 17:00:19  joe
-- time /geht nicht
--
-- Revision 1.6  2002/08/31 13:44:16  joe
-- args -> args'

import Main.Socketeer
import qualified World

import Util.Debug
import Util.Args
import Util.Call
import Util.Command
import Util.Value
import Util.FiniteMap

import Main.Simple
import Patti.Patti2

import Main.Timed

import World.Type

import Geo.Segmentation
import Geo.Printer

import System
import IO
import IOExts
import Monad (when)
import CPUTime

import Control.Concurrent (forkIO, threadDelay, killThread, ThreadId)
import Control.Concurrent.Chan
import Control.Concurrent.MVar

default_host = "icfp1.cse.ogi.edu"
default_port = "22010"


main :: IO ()
main = do
    argv <- getArgs
    case argv of
	 [ host ] -> run         host default_port
	 [ host, port ] -> run   host         port
	 _ ->        run default_host default_port


run :: String -> String -> IO ()
run host port = do
    h <- dial host port
    hput h "Player"
    
    (d, (m, i)) <- timed $ World.create h
    debug $ "... World created, time: " ++ show d

    (d, ( dist , graph )) <- timed $ segment_by_homes (m, i)
    -- print_dmap i dist
    -- print_wgraph graph
    debug $ "voronoi time: " ++ show d

    hwf <- newIORef ( 0, emptyFM )
    loop $ Args { handle = h
		, mappe = m
		, info = i
		, waygraph = graph
		, distancemap = dist
		, highwaysfrom = hwf

		, time_per_move = 10 ^ 6

		, number = 1
		}

loop :: Args -> IO ()
loop args = do

    debug $ "number = " ++ show (number args)
    t <- getCPUTime
    let tms = fromIntegral $ t `div` (10^9)
    let avg = tms `div` (number args)
    debug $ unwords
	  [ "CPU time already used:", show t, "\n"
	  , "rounded to", show tms, "milliseconds", "\n" 
	  , "average: ", show avg, "milliseconds per move."
	  ]
    
    when ( avg > 1000 ) $ debug "WARNING: avg time > 1000 ms"


    msg <- hget $ handle args
    i <- return $ World.update ( info args ) msg
    debug $ "new info: " ++ show i
    args <- return $ args { info = i }

    msg <- hget $ handle args
    i <- World.pack_update ( mappe args , info args ) msg
    debug $ "new info (pack_update): " ++ show i
    args <- return $ args { info = i }

    command <- generate args suggestors
    i <- return $ World.money_update ( info args ) ( abs $ bid command )
    args <- return $ args { info = i }

    hput ( handle args ) $ show command
    
    loop $ args { number = succ $ number $ args }
    

suggestors :: [ Suggestor ]
suggestors = [ simple , patti2 ]

generate :: Args -> [ Suggestor ] ->  IO Command
generate args suggs = do

    ch <- newChan :: IO (Chan (Maybe (Value, Command)))

    -- used by the clients
    let acc x = writeChan ch $ Just x

    -- sent from the killer timer, and kills everything
    let kill = writeChan ch $ Nothing

    debug "fork suggestors ..."
    pids <- mapM ( launch args acc ) suggs    -- everything is forked
    debug "... suggestors forked"

    forkIO $ do
        threadDelay $ time_per_move args  -- clock ticks in real micro-seconds ??
        kill

    top <- newEmptyMVar :: IO (MVar (Value, Command))

    waiter args ch top 
    killall pids

    mtop <- tryTakeMVar top    
    case mtop of
	 Just ( val, com ) -> do
	    debug $ "picking best suggestion"
	    return com
	 Nothing -> do
	    debug $ "oops - no suggestions! sending fallback"
	    return $ Command { bid = 1 , action = Pick [] }
        
killall :: [ ThreadId ] -> IO ()
killall pids = do
    debug "killing all suggestors ..."
    mapM_ killThread pids `catch` \ _ -> return ()
    debug "... killed all suggestors"


launch :: Args -> Acceptor -> Suggestor -> IO ThreadId
launch args acc sugg = do
    forkIO $  sugg args acc

    

waiter :: Args -> Chan (Maybe (Value, Command)) -> MVar  (Value, Command) 
       ->  IO ()
waiter args ch top = do
    mitem <- readChan ch
    debug $ "first item: " ++ show mitem
    case mitem of
	 Nothing -> debug $ "Nothing? anyway."
	 Just item -> do
	      putMVar top item
	      collector args ch top

collector :: Args -> Chan (Maybe (Value, Command)) -> MVar  (Value, Command) 
       ->  IO ()
collector args ch top = do
    mitem <- readChan ch
    debug $ "next item: " ++ show mitem
    case mitem of
	 Nothing -> debug $ "ok, time is over"
	 Just (this @ (v, c)) -> do
	      prev <- takeMVar top
	      putMVar top (max this prev)
	      if (v >= max_value) 
		 then debug $ "this is max_value, returning."
		 else do
		      debug $ "this NOT max_value, continue"
		      collector args ch top




    
	     

    


