module World.Type where

import Data.Word
import Data.Array.IO
import Util.RestrictedFiniteMap
import Util.RestrictedSet
import Util.Direction

------------------------------------------------------------------------------

-- direct addressing mode
-- contiguous items
type Pos = Int

type Map                      =  IOUArray Pos Word8
type RobotID                  =  Int
type PackID                   =  Int

-- high-level addressing mode
-- two-dimensional items
type Position                 =  ( Int , Int )


------------------------------------------------------------------------------

data Package                  =  Package
    { weight                  :: Maybe Int
    , destination             :: Maybe Position
    , pos                     :: Either Position RobotID 
    , was_dropped             :: Bool
    }                            deriving ( Eq , Ord , Show )

------------------------------------------------------------------------------

data Bool3                    =  True3 | False3 | Unknown3
			      deriving ( Eq , Ord, Show )

data Info                     =  Info
    { my_id                   :: RobotID
    , my_capacity             :: Int
    , my_money                :: Int
    , my_score                :: Int
    , field_size              :: Position
    , packages                :: FiniteMap PackID   Package
    , robot_packs             :: FiniteMap RobotID ( Set PackID )
    , robot_poss              :: FiniteMap RobotID  Position
    , our_actions             :: [ Action ]
    , hb_contents             :: FiniteMap Position Bool3
    }                            deriving ( Eq )

instance                         Show Info where
    show ( Info i c m s _ _ rpas rpos _ _ )
	                      =  unlines $
				 [ "id = " ++ show i 
				   ++ ", capacity = " ++ show c
				   ++ ", money = " ++ show m
				   ++ ", score = " ++ show s
				 , "packs = " ++ show rpas
				 , "poss = " ++ show rpos
				 ]

instance ( Show a , Show b )  => Show ( FiniteMap a b ) where
    show                      =  show . fmToList

instance ( Show a ) => Show (Set a) where
    show            = show . setToList

-----------------------------------------------------------------------------

max_actions_in_history        :: Int
max_actions_in_history        =  50

add_action                    :: Action -> [ Action ] -> [ Action ]
add_action a as               =  take max_actions_in_history $ a : as

-----------------------------------------------------------------------------

data Field                    =  Water
			      |  Open
			      |  Wall
			      |  Base
				 deriving ( Eq , Ord , Show )

-----------------------------------------------------------------------------

type Bid = Int

data Action = Move Direction
	    | Pick [ PackID ]
	    | Drop [ PackID ]
     deriving ( Eq, Ord )

instance Show Action where
    show ( Move d ) = unwords [ "Move" , show d ]
    show ( Pick packs ) = unwords $ "Pick" : map show packs
    show ( Drop packs ) = unwords $ "Drop" : map show packs
